/* Low-level functions for atomic operations.  PA-RISC version. -*- C++ -*-
   Copyright 2001 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#ifndef _BITS_ATOMICITY_H
#define _BITS_ATOMICITY_H	1

/* Define this to 1 when everyone is using a sensible glibc.  */
#undef MALLOC_ALIGNS_TO_16BYTE

/* Load and clear, the only PA-RISC atomic read-write operation.  Some
   cpus only support ldcw on 16 byte aligned words.  *sigh*.  */
static inline int
__pa_ldcw (volatile int* lock)
{
  int ret;
  __asm__ __volatile__ ("ldcw 0(%1),%0" : "=r" (ret) : "r" (lock));
  return ret;
}

static inline int
__pa_spin_lock (volatile int* lock)
{
  int ret;
  while ((ret = __pa_ldcw (lock)) == 0)
    /* If we don't grab the lock, don't clobber the bus with ldcw.  */
    while (*lock == 0) /* spin */ ;
  return ret;
}


class _Atomic_counter {
 public:
  int lock;
  int count;
#if ! MALLOC_ALIGNS_TO_16BYTE
  /* Two locks???  Well, malloc allocates to 8 byte boundaries, and
     thus either lock or alt_lock will be 16 byte aligned.  Use
     whichever one gets the right alignment.  */
  int alt_lock;
#endif

  inline
  _Atomic_counter()
  {
    lock = 1;
#if ! MALLOC_ALIGNS_TO_16BYTE
    alt_lock = 1;
#endif
  }

  inline int
  operator=(const int __val) { return this->count = __val; }

  inline bool
  operator<(const int __rhs) const
    { return this->count < __rhs; }

  inline bool
  operator>(const int __rhs) const
    { return this->count > __rhs; }
};

typedef class _Atomic_counter _Atomic_word;

/* This is horrible, but how else can we get the static initializer
   correct with our lock(s) set non-zero? */
#if MALLOC_ALIGNS_TO_16BYTE
#define _S_EMPTY_REP_INIT = { 0, 0, 1, 0, 0 }
#else
#define _S_EMPTY_REP_INIT = { 0, 0, 1, 0, 1, 0 }
#endif


static inline int
__attribute__ ((__unused__))
__exchange_and_add (_Atomic_word* mem, int val)
{
  int result;
  int tmp;
#if MALLOC_ALIGNS_TO_16BYTE
  int *lock = &mem->lock;
#else
  int *lock = (int *) ((long) &mem->alt_lock & ~15);
#endif

  tmp = __pa_spin_lock (lock);
  result = mem->count;
  mem->count += val;
  __asm__ __volatile__("");
  *lock = tmp;
  return result;
}

static inline void
__attribute__ ((__unused__))
__atomic_add (_Atomic_word* mem, int val)
{
  int tmp;
#if MALLOC_ALIGNS_TO_16BYTE
  int *lock = &mem->lock;
#else
  int *lock = (int *) ((long) &mem->alt_lock & ~15);
#endif

  tmp = __pa_spin_lock (lock);
  mem->count += val;
  __asm__ __volatile__("");
  *lock = tmp;
}

#endif
